function [S,U,V,W,info] = btd_grad_full(A,S,U,V,W,tol,itermax)
% [S,U,V,W,info] = btd_grad_full(A,S,U,V,W,tol,itermax) tries to compute the best
% approximation (in the least-squares sense) of a third-order tensor by a BTD
% using the Riemannian gradient algorithm with backtracking and without variable projection.
% INPUT:
%  - A is the third-order tensor to be approximated by BTD;
%  - S is a fourth-order tensor containing the initial (third-order) core
%    tensors S(:,:,:,1),...,S(:,:,:,R);
%  - U, V and W are third-order tensors containing the initial factor matrices
%    U(:,:,1),...,U(:,:,R), V(:,:,1),...,V(:,:,R) and W(:,:,1),...,W(:,:,R);
%  - tol is the tolerance on the norm of the gradient;
%  - itermax is the maximum number of iterations in the main loop.
% OUTPUT:
%  - S is a fourth-order tensor containing the (third-order) core tensors
%    S(:,:,:,1),...,S(:,:,:,R);
%  - U, V and W are third-order tensors containing the factor matrices
%    U(:,:,1),...,U(:,:,R), V(:,:,1),...,V(:,:,R) and W(:,:,1),...,W(:,:,R);
%  - info is a structure gathering useful data about the run: the number
%    of iterations in each backtracking loop ('iterations'), the norm of the
%    gradient at each iteration ('normgrad'), the value of the objective function
%    at each iteration ('objective').
%% Sizes
I = size(A);
Rank = zeros(1,3);
[Rank(1),Rank(2),Rank(3),R] = size(S); % Rank is the multilinear rank of each term in the BTD
% R is the number of terms in the BTD
%% Step length: paramaters for backtracking
itermax2 = 10; % maximum number of iterations in the backtracking loop
a = 0.4;
b = 0.5;
c = 1e-3;
%% Initialization
% iteration counters
iter = 1; % main loop
iter2(1) = 0; % inner loop for backtracking
% iterates
for r = 1:R
    U(:,:,r) = ProjStiefel(U(:,:,r));
    V(:,:,r) = ProjStiefel(V(:,:,r));
    W(:,:,r) = ProjStiefel(W(:,:,r));
end
% objective
testObj = f(A,I,R,S,U,V,W);
obj(1) = testObj;
% gradient
[gS,gU,gV,gW] = grad_f(A,I,R,Rank,S,U,V,W);
normgrad(1) = sqrt(frob(gS)^2+norm(gU,'fro')^2+norm(gV,'fro')^2+norm(gW,'fro')^2);
% test iterates
testS = zeros([Rank R]);
testU = zeros(I(1),Rank(1),R);
testV = zeros(I(2),Rank(2),R);
testW = zeros(I(3),Rank(3),R);
%% Loop
while normgrad(end) > tol && iter < itermax
    iter2(iter) = 0;
    % step length selection
    t = a;
    target = obj(iter)-c*t*normgrad(iter)^2;
    while testObj > target && iter2(iter) < itermax2
        t = t*b;
        target = obj(iter)-c*t*normgrad(iter)^2;
        testS = S-t*gS;
        for r = 1:R % retraction
            testU(:,:,r) = qf(U(:,:,r)-t*gU(:,(r-1)*Rank(1)+1:r*Rank(1)));
            testV(:,:,r) = qf(V(:,:,r)-t*gV(:,(r-1)*Rank(2)+1:r*Rank(2)));
            testW(:,:,r) = qf(W(:,:,r)-t*gW(:,(r-1)*Rank(3)+1:r*Rank(3)));
        end
        testObj = f(A,I,R,testS,testU,testV,testW);
        iter2(iter) = iter2(iter)+1;
    end
    % update of the iterates
    S = testS;
    U = testU;
    V = testV;
    W = testW;
    [gS,gU,gV,gW] = grad_f(A,I,R,Rank,S,U,V,W);
    iter = iter+1;
    obj(iter) = testObj;
    normgrad(iter) = sqrt(frob(gS)^2+norm(gU,'fro')^2+norm(gV,'fro')^2+norm(gW,'fro')^2);
end
info = struct('iterations',iter2,'normgrad',normgrad,'objective',obj);
end